/*
 * Copyright  2004 David Reveman
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the names of
 * David Reveman not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * David Reveman makes no representations about the suitability of this
 * software for any purpose. It is provided "as is" without express or
 * implied warranty.
 *
 * DAVID REVEMAN DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
 * NO EVENT SHALL DAVID REVEMAN BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: David Reveman <c99drn@cs.umu.se>
 */

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include "glitzint.h"

#define SET_COLOR(color, set_red, set_green, set_blue, set_alpha) \
  (color).red = (set_red); \
  (color).green = (set_green); \
  (color).blue = (set_blue); \
  (color).alpha = (set_alpha)

glitz_gl_uint_t _texture_units[] = {
  GLITZ_GL_TEXTURE0,
  GLITZ_GL_TEXTURE1,
  GLITZ_GL_TEXTURE2
};

typedef struct _glitz_texture_unit_t {
  glitz_texture_t *texture;
  glitz_gl_uint_t unit;
  glitz_bool_t transform;
} glitz_texture_unit_t;

void
glitz_composite (glitz_operator_t op,
                 glitz_surface_t *src,
                 glitz_surface_t *mask,
                 glitz_surface_t *dst,
                 int x_src,
                 int y_src,
                 int x_mask,
                 int y_mask,
                 int x_dst,
                 int y_dst,
                 int width,
                 int height)
{
  glitz_composite_op_t comp_op;
  int i, passes, texture_nr = -1;
  glitz_texture_t *stexture, *mtexture;
  glitz_texture_unit_t textures[3];
  glitz_color_t alpha_mask;
  glitz_gl_enum_t primitive;
  glitz_gl_int_t first;
  glitz_gl_sizei_t count;
  glitz_box_t rect;

  GLITZ_GL_SURFACE (dst);

  if (width <= 0 || height <= 0)
    return;
  
  glitz_composite_op_init (&comp_op, op, src, mask, dst);
  if (comp_op.type == GLITZ_COMBINE_TYPE_NA) {
    glitz_surface_status_add (dst, GLITZ_STATUS_NOT_SUPPORTED_MASK);
    return;
  }

  src = comp_op.src;
  mask = comp_op.mask;

  if (src) {
    stexture = glitz_surface_get_texture (src, 0);
    if (!stexture)
      return;
  } else
    stexture = NULL;

  if (mask) {
    mtexture = glitz_surface_get_texture (mask, 0);
    if (!mtexture)
      return;
  } else
    mtexture = NULL;

  if (!glitz_surface_push_current (dst, GLITZ_DRAWABLE_CURRENT)) {
    glitz_surface_status_add (dst, GLITZ_STATUS_NOT_SUPPORTED_MASK);
    glitz_surface_pop_current (dst);
    return;
  }

  rect.x1 = x_dst;
  rect.y1 = y_dst;
  rect.x2 = rect.x1 + width;
  rect.y2 = rect.y1 + height;

  if (mtexture) {
    textures[0].texture = mtexture;
    textures[0].unit = _texture_units[0];
    textures[0].transform = 0;
    texture_nr = 0;

    glitz_texture_bind (gl, mtexture);

    glitz_texture_set_tex_gen (gl,
                               mtexture,
                               rect.x1 - x_mask,
                               rect.y1 - y_mask,
                               mask->flags);
    
    if (mask->transform) {
      textures[0].transform = 1;
      gl->matrix_mode (GLITZ_GL_TEXTURE);
      gl->load_matrix_f (SURFACE_EYE_COORDS (mask)?
                         mask->transform->m: mask->transform->t);
      gl->matrix_mode (GLITZ_GL_MODELVIEW);
      
      if (SURFACE_LINEAR_TRANSFORM_FILTER (mask))
        glitz_texture_ensure_filter (gl, mtexture, GLITZ_GL_LINEAR);
      else
        glitz_texture_ensure_filter (gl, mtexture, GLITZ_GL_NEAREST);
    } else
      glitz_texture_ensure_filter (gl, mtexture, GLITZ_GL_NEAREST);
    
    if (SURFACE_REPEAT (mask)) {
      if (SURFACE_MIRRORED (mask))
        glitz_texture_ensure_wrap (gl, mtexture, GLITZ_GL_MIRRORED_REPEAT);
      else
        glitz_texture_ensure_wrap (gl, mtexture, GLITZ_GL_REPEAT);
    } else {
      if ((!(dst->drawable->backend->feature_mask &
             GLITZ_FEATURE_TEXTURE_BORDER_CLAMP_MASK)) ||
          SURFACE_PAD (mask))
        glitz_texture_ensure_wrap (gl, mtexture, GLITZ_GL_CLAMP_TO_EDGE);
      else
        glitz_texture_ensure_wrap (gl, mtexture, GLITZ_GL_CLAMP_TO_BORDER);
    }
  }
    
  if (stexture) {
    int last_texture_nr = comp_op.combine->texture_units - 1;
    
    while (texture_nr < last_texture_nr) {
      textures[++texture_nr].texture = stexture;
      textures[texture_nr].unit = _texture_units[texture_nr];
      textures[texture_nr].transform = 0;
      if (texture_nr > 0)
        gl->active_texture (textures[texture_nr].unit);
      glitz_texture_bind (gl, stexture);
    }

    glitz_texture_set_tex_gen (gl,
                               stexture,
                               rect.x1 - x_src,
                               rect.y1 - y_src,
                               src->flags);

    if (src->transform) {
      textures[texture_nr].transform = 1;
      gl->matrix_mode (GLITZ_GL_TEXTURE);
      gl->load_matrix_f (SURFACE_EYE_COORDS (src)?
                         src->transform->m: src->transform->t);
      gl->matrix_mode (GLITZ_GL_MODELVIEW);

      if (SURFACE_LINEAR_TRANSFORM_FILTER (src))
        glitz_texture_ensure_filter (gl, stexture, GLITZ_GL_LINEAR);
      else
        glitz_texture_ensure_filter (gl, stexture, GLITZ_GL_NEAREST);
    } else
      glitz_texture_ensure_filter (gl, stexture, GLITZ_GL_NEAREST);
      
    if (SURFACE_REPEAT (src)) {
      if (SURFACE_MIRRORED (src))
        glitz_texture_ensure_wrap (gl, stexture, GLITZ_GL_MIRRORED_REPEAT);
      else
        glitz_texture_ensure_wrap (gl, stexture, GLITZ_GL_REPEAT);
    } else {
      if ((!(dst->drawable->backend->feature_mask &
             GLITZ_FEATURE_TEXTURE_BORDER_CLAMP_MASK)) ||
          SURFACE_PAD (src))
        glitz_texture_ensure_wrap (gl, stexture, GLITZ_GL_CLAMP_TO_EDGE);
      else
        glitz_texture_ensure_wrap (gl, stexture, GLITZ_GL_CLAMP_TO_BORDER);
    }
  }

  gl->scissor (rect.x1 + dst->x,
               dst->attached->height - dst->y - rect.y2,
               width, height);

  gl->push_matrix ();

  glitz_geometry_enable (gl, dst, &primitive, &first, &count, &rect);

  if (dst->indirect) {
    glitz_sample_offset_t *offsets;

    passes = dst->indirect->n_samples;
    offsets = dst->indirect->offsets;

    gl->enable (GLITZ_GL_STENCIL_TEST);
    gl->disable (GLITZ_GL_BLEND);
    gl->color_mask (GLITZ_GL_FALSE, GLITZ_GL_FALSE, GLITZ_GL_FALSE,
                    GLITZ_GL_FALSE);
    gl->clear_stencil (0);
    gl->clear (GLITZ_GL_STENCIL_BUFFER_BIT);
    gl->stencil_func (GLITZ_GL_ALWAYS, ~0, ~0);
    gl->stencil_op (GLITZ_GL_KEEP, GLITZ_GL_INCR, GLITZ_GL_INCR);
    
    for (i = 0; i < passes; i++) {
      gl->push_matrix ();
      gl->translate_f (offsets[i].x, offsets[i].y, 0.0f);
      gl->draw_arrays (primitive, first, count);
      gl->pop_matrix ();
    }

    gl->pop_matrix ();
    gl->push_matrix ();
    
    gl->color_mask (GLITZ_GL_TRUE, GLITZ_GL_TRUE, GLITZ_GL_TRUE,
                    GLITZ_GL_TRUE);

    glitz_geometry_disable (gl, dst);
    glitz_geometry_enable_default (gl, dst, &rect);
    
    primitive = GLITZ_GL_QUADS;
    first = 0;
    count = 4;
  } else
    passes = 1;

  if (comp_op.per_component)
    passes *= comp_op.per_component;

  alpha_mask = comp_op.alpha_mask;

  for (i = 0; i < passes; i++) {
    unsigned short weight, sample;
    
    if (dst->indirect) {
      if (comp_op.per_component)
        sample = i / comp_op.per_component;
      else
        sample = i;
      
      weight = dst->indirect->weights[sample];
      
      gl->stencil_func (GLITZ_GL_EQUAL, sample + 1, ~0);
      gl->stencil_op (GLITZ_GL_KEEP, GLITZ_GL_KEEP, GLITZ_GL_KEEP);
    } else
      weight = 0xffff;
    
    if (comp_op.per_component) {
      switch (i % comp_op.per_component) {
      case 0:
        SET_COLOR (comp_op.alpha_mask,
                   SHORT_MULT (weight, alpha_mask.red), 0, 0, 0);
        gl->color_mask (GLITZ_GL_TRUE, GLITZ_GL_FALSE, GLITZ_GL_FALSE,
                        GLITZ_GL_FALSE);
        break;
      case 1:
        SET_COLOR (comp_op.alpha_mask,
                   0, 0, SHORT_MULT (weight, alpha_mask.blue), 0);
        gl->color_mask (GLITZ_GL_FALSE, GLITZ_GL_FALSE, GLITZ_GL_TRUE,
                        GLITZ_GL_FALSE);
        break;
      case 2:
        SET_COLOR (comp_op.alpha_mask,
                   0, SHORT_MULT (weight, alpha_mask.green), 0, 0);
        gl->color_mask (GLITZ_GL_FALSE, GLITZ_GL_TRUE, GLITZ_GL_FALSE,
                        GLITZ_GL_FALSE);
        break;
      case 3:
        SET_COLOR (comp_op.alpha_mask,
                   0, 0, 0, SHORT_MULT (weight, alpha_mask.alpha));
        gl->color_mask (GLITZ_GL_FALSE, GLITZ_GL_FALSE, GLITZ_GL_FALSE,
                        GLITZ_GL_TRUE);
        break;
      }
    } else {
      SET_COLOR (comp_op.alpha_mask,
                 SHORT_MULT (weight, alpha_mask.red),
                 SHORT_MULT (weight, alpha_mask.green),
                 SHORT_MULT (weight, alpha_mask.blue),
                 SHORT_MULT (weight, alpha_mask.alpha));
    }
  
    glitz_composite_enable (&comp_op);
    
    gl->draw_arrays (primitive, first, count);
  }

  glitz_geometry_disable (gl, dst);

  gl->pop_matrix ();

  if (dst->indirect)
    gl->disable (GLITZ_GL_STENCIL_TEST);
  
  if (comp_op.per_component)
    gl->color_mask (GLITZ_GL_TRUE, GLITZ_GL_TRUE, GLITZ_GL_TRUE,
                    GLITZ_GL_TRUE);
  
  glitz_composite_disable (&comp_op);

  for (i = texture_nr; i >= 0; i--) {
    glitz_texture_unbind (gl, textures[i].texture);
    if (textures[i].transform) {
      gl->matrix_mode (GLITZ_GL_TEXTURE);
      gl->load_identity ();
      gl->matrix_mode (GLITZ_GL_MODELVIEW);
    }

    if (i > 0)
      gl->active_texture (textures[i - 1].unit);
  }

  glitz_surface_damage (dst, &rect,
                        GLITZ_DAMAGE_TEXTURE_MASK |
                        GLITZ_DAMAGE_SOLID_MASK);
  
  glitz_surface_pop_current (dst);
}

void
glitz_copy_area (glitz_surface_t *src,
                 glitz_surface_t *dst,
                 int x_src,
                 int y_src,
                 int width,
                 int height,
                 int x_dst,
                 int y_dst)
{
  glitz_status_t status;

  GLITZ_GL_SURFACE (dst);
  
  if (x_src < 0) {
    x_dst -= x_src;
    width += x_src;
    x_src = 0;
  }

  if (y_src < 0) {
    y_dst -= y_src;
    height += y_src;
    y_src = 0;
  }

  width = MIN (src->width - x_src, width);
  height = MIN (src->height - y_src, height);

  if (x_dst < 0) {
    x_src -= x_dst;
    width += x_dst;
    x_dst = 0;
  }

  if (y_dst < 0) {
    y_src -= y_dst;
    height += y_dst;
    y_dst = 0;
  }
  
  width = MIN (dst->width - x_dst, width);
  height = MIN (dst->height - y_dst, height);

  if (width <= 0 || height <= 0)
    return;

  status = GLITZ_STATUS_NOT_SUPPORTED;
  if (glitz_surface_push_current (dst, GLITZ_DRAWABLE_CURRENT)) {
    glitz_box_t box;
    
    if (src->attached == dst->attached) {
      if (REGION_NOTEMPTY (&src->drawable_damage)) {
        glitz_surface_push_current (src, GLITZ_DRAWABLE_CURRENT);
        glitz_surface_pop_current (src);
      }
      status = GLITZ_STATUS_SUCCESS;
    } else
      status = glitz_surface_make_current_read (src);

    box.x1 = x_dst;
    box.y1 = y_dst;
    box.x2 = box.x1 + width;
    box.y2 = box.y1 + height;

    if (!status) {
      gl->read_buffer (src->buffer);
      gl->draw_buffer (dst->buffer);
      
      glitz_set_operator (gl, GLITZ_OPERATOR_SRC);
      
      gl->disable (GLITZ_GL_SCISSOR_TEST);
      
      glitz_set_raster_pos (gl,
                            x_dst + dst->x,
                            dst->attached->height - (y_dst + dst->y + height));
      gl->copy_pixels (x_src + src->x,
                       src->attached->height - (y_src + src->y + height),
                       width, height, GLITZ_GL_COLOR);

      gl->enable (GLITZ_GL_SCISSOR_TEST);
    } else {
      glitz_texture_t *texture;

      texture = glitz_surface_get_texture (src, 0);
      if (texture) {
        glitz_texture_bind (gl, texture);

        glitz_texture_set_tex_gen (gl, texture,
                                   x_dst - x_src,
                                   y_dst - y_src,
                                   0);

        gl->tex_env_f (GLITZ_GL_TEXTURE_ENV, GLITZ_GL_TEXTURE_ENV_MODE,
                       GLITZ_GL_REPLACE);
        gl->color_4us (0x0, 0x0, 0x0, 0xffff);
        
        glitz_texture_ensure_wrap (gl, texture, GLITZ_GL_CLAMP_TO_EDGE);
        glitz_texture_ensure_filter (gl, texture, GLITZ_GL_NEAREST);

        glitz_set_operator (gl, GLITZ_OPERATOR_SRC);

        gl->scissor (box.x1 + dst->x,
                     dst->attached->height - dst->y - box.y2,
                     width, height);
        
        glitz_geometry_enable_default (gl, dst, &box);
        
        gl->draw_arrays (GLITZ_GL_QUADS, 0, 4);

        glitz_texture_unbind (gl, texture);
      }
    }
    
    glitz_surface_damage (dst, &box,
                          GLITZ_DAMAGE_TEXTURE_MASK |
                          GLITZ_DAMAGE_SOLID_MASK);

    status = GLITZ_STATUS_SUCCESS;
  }

  glitz_surface_pop_current (dst);

  if (status) {
    if (glitz_surface_push_current (src, GLITZ_DRAWABLE_CURRENT)) {
      glitz_texture_t *texture;

      gl->read_buffer (src->buffer);

      texture = glitz_surface_get_texture (dst, 1);
      if (texture) {
        glitz_box_t box;

        gl->disable (GLITZ_GL_SCISSOR_TEST);

        glitz_texture_bind (gl, texture);

        glitz_texture_copy_drawable (gl,
                                     texture,
                                     src->attached,
                                     x_src + src->x,
                                     y_src + src->y,
                                     width, height,
                                     x_dst, y_dst);

        glitz_texture_unbind (gl, texture);

        gl->enable (GLITZ_GL_SCISSOR_TEST);

        box.x1 = x_dst;
        box.y1 = y_dst;
        box.x2 = box.x1 + width;
        box.y2 = box.y1 + height;

        glitz_surface_damage (dst, &box,
                              GLITZ_DAMAGE_DRAWABLE_MASK |
                              GLITZ_DAMAGE_SOLID_MASK);
        
        status = GLITZ_STATUS_SUCCESS;
      }
    }
    glitz_surface_pop_current (src);
  }

  if (status)
    glitz_surface_status_add (dst, glitz_status_to_status_mask (status));
}
